from __future__ import annotations

import re
from typing import Literal

import numpy as np
import pytest
import rasterio as rio
from scipy.interpolate import interpn
from scipy.ndimage import binary_dilation

import geoutils as gu
from geoutils import examples
from geoutils.interface.interpolate import (
    _get_dist_nodata_spread,
    _interp_points,
    _interpn_interpolator,
    method_to_order,
)
from geoutils.projtools import reproject_to_latlon


class TestInterpolate:

    landsat_b4_path = examples.get_path_test("everest_landsat_b4")
    aster_dem_path = examples.get_path_test("exploradores_aster_dem")
    landsat_b4_crop_path = examples.get_path_test("everest_landsat_b4_cropped")
    landsat_rgb_path = examples.get_path_test("everest_landsat_rgb")

    def test_dist_nodata_spread(self) -> None:
        """Test distance of nodata spreading computation based on interpolation order."""

        assert _get_dist_nodata_spread(0, "half_order_up") == 0
        assert _get_dist_nodata_spread(0, "half_order_down") == 0
        assert _get_dist_nodata_spread(0, 5) == 5

        assert _get_dist_nodata_spread(1, "half_order_up") == 1
        assert _get_dist_nodata_spread(1, "half_order_down") == 0
        assert _get_dist_nodata_spread(1, 5) == 5

        assert _get_dist_nodata_spread(3, "half_order_up") == 2
        assert _get_dist_nodata_spread(3, "half_order_down") == 1
        assert _get_dist_nodata_spread(3, 5) == 5

        assert _get_dist_nodata_spread(5, "half_order_up") == 3
        assert _get_dist_nodata_spread(5, "half_order_down") == 2
        assert _get_dist_nodata_spread(5, 5) == 5

    @pytest.mark.parametrize(
        "method", ["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"]
    )  # type: ignore
    def test_interpn_interpolator_accuracy(
        self, method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"]
    ) -> None:
        """Test that _interpn_interpolator (used by interp_points) returns exactly the same result as scipy.interpn."""

        # Create synthetic 2D array with non-aligned coordinates in X/Y, and X in descending order to mirror a raster's
        # coordinates
        shape = (50, 20)
        coords = (np.linspace(10, 0, shape[0]), np.linspace(20, 30, shape[1]))
        values = np.random.default_rng(42).normal(size=shape)

        # Get 10 random points in the array boundaries
        i = np.random.default_rng(42).uniform(coords[0][-1], coords[0][0], size=10)
        j = np.random.default_rng(42).uniform(coords[1][0], coords[1][-1], size=10)

        # Compare interpn and interpolator

        # Method splinef2d is expecting strictly ascending coordinates (while other methods support desc or asc)
        if method != "splinef2d":
            vals = interpn(points=coords, values=values, xi=(i, j), method=method)
        else:
            vals = interpn(
                points=(np.flip(coords[0]), coords[1]), values=np.flip(values[:], axis=0), xi=(i, j), method=method
            )
        # With the interpolator (coordinates are re-ordered automatically, as it happens often for rasters)
        interpolator = _interpn_interpolator(points=coords, values=values, method=method)
        vals2 = interpolator((i, j))

        assert np.array_equal(vals, vals2, equal_nan=True)

    @pytest.mark.parametrize("tag_aop", [None, "Area", "Point"])  # type: ignore
    @pytest.mark.parametrize("shift_aop", [True, False])  # type: ignore
    def test_interp_points__synthetic(self, tag_aop: Literal["Area", "Point"] | None, shift_aop: bool) -> None:
        """
        Test interp_points function with synthetic data:

        We select known points and compare to the expected interpolation results across all methods, and all pixel
        interpretations (area_or_point, with or without shift).

        The synthetic data is a 3x3 array of values, which we interpolate at different synthetic points:
        1/ Points falling right on grid coordinates are compared to their value in the grid,
        2/ Points falling halfway between grid coordinates are compared to the mean of the surrounding values in the
        grid (as it should equal the linear interpolation on an equal grid),
        3/ Random points in the grid, compared between methods (forcing to use either scipy.ndimage.map_coordinates or
        scipy.interpolate.interpn under-the-hood, to ensure results are consistent).

        These tests also check the behaviour when returning interpolated for points outside of valid values.
        """

        # We flip the array up/down to facilitate index comparison of Y axis
        arr = np.flipud(np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]).reshape((3, 3)))
        transform = rio.transform.from_bounds(0, 0, 3, 3, 3, 3)
        raster = gu.Raster.from_array(data=arr, transform=transform, crs=None, nodata=-9999)

        # Define the AREA_OR_POINT attribute without re-transforming
        raster.set_area_or_point(tag_aop, shift_area_or_point=False)

        # Check interpolation falls right on values for points (1, 1), (1, 2) etc...
        index_x = [0, 1, 2, 0, 1, 2, 0, 1, 2]
        index_y = [0, 0, 0, 1, 1, 1, 2, 2, 2]

        # The actual X/Y coords will be offset by one because Y axis is inverted and pixel coords is upper-left corner
        points_x, points_y = raster.ij2xy(i=index_x, j=index_y, shift_area_or_point=shift_aop)

        # The following 4 methods should yield the same result because:
        # Nearest = Linear interpolation at the location of a data point
        # Regular grid = Equal grid interpolation at the location of a data point

        raster_points = raster.interp_points(
            (points_x, points_y), method="nearest", shift_area_or_point=shift_aop, as_array=True
        )
        raster_points_lin = raster.interp_points(
            (points_x, points_y), method="linear", shift_area_or_point=shift_aop, as_array=True
        )
        raster_points_interpn = raster.interp_points(
            (points_x, points_y),
            method="nearest",
            force_scipy_function="interpn",
            shift_area_or_point=shift_aop,
            as_array=True,
        )
        raster_points_interpn_lin = raster.interp_points(
            (points_x, points_y),
            method="linear",
            force_scipy_function="interpn",
            shift_area_or_point=shift_aop,
            as_array=True,
        )

        assert np.array_equal(raster_points, raster_points_lin)
        assert np.array_equal(raster_points, raster_points_interpn)
        assert np.array_equal(raster_points, raster_points_interpn_lin)

        for i in range(3):
            for j in range(3):
                ind = 3 * i + j
                assert raster_points[ind] == arr[index_x[ind], index_y[ind]]

        # Check bilinear interpolation values inside the grid (same here, offset by 1 between X and Y)
        index_x_in = [0.5, 0.5, 1.5, 1.5]
        index_y_in = [0.5, 1.5, 0.5, 1.5]

        points_x_in, points_y_in = raster.ij2xy(i=index_x_in, j=index_y_in, shift_area_or_point=shift_aop)

        # Here again compare methods
        raster_points_in = raster.interp_points(
            (points_x_in, points_y_in), method="linear", shift_area_or_point=shift_aop, as_array=True
        )
        raster_points_in_interpn = raster.interp_points(
            (points_x_in, points_y_in),
            method="linear",
            force_scipy_function="interpn",
            shift_area_or_point=shift_aop,
            as_array=True,
        )

        assert np.array_equal(raster_points_in, raster_points_in_interpn)

        for i in range(len(points_x_in)):
            xlow = int(index_x_in[i] - 0.5)
            xupp = int(index_x_in[i] + 0.5)
            ylow = int(index_y_in[i] - 0.5)
            yupp = int(index_y_in[i] + 0.5)

            # Check the bilinear interpolation matches the mean value of those 4 points (equivalent as its the middle)
            assert raster_points_in[i] == np.mean([arr[xlow, ylow], arr[xupp, ylow], arr[xupp, yupp], arr[xlow, yupp]])

        # Check bilinear extrapolation for points at 1 spacing outside from the input grid
        points_out = (
            [(-1, i) for i in np.arange(1, 4)]
            + [(i, -1) for i in np.arange(1, 4)]
            + [(4, i) for i in np.arange(1, 4)]
            + [(i, 4) for i in np.arange(4, 1)]
        )
        points_out_xy = list(zip(*points_out))
        raster_points_out = raster.interp_points(points_out_xy, as_array=True)
        assert all(~np.isfinite(raster_points_out))

        # To use cubic or quintic, we need a larger grid (minimum 6x6, but let's aim bigger with 50x50)
        arr = np.flipud(np.arange(1, 2501).reshape((50, 50)))
        transform = rio.transform.from_bounds(0, 0, 50, 50, 50, 50)
        raster = gu.Raster.from_array(data=arr, transform=transform, crs=None, nodata=-9999)
        raster.set_area_or_point(tag_aop, shift_area_or_point=False)

        # For this, get random points
        rng = np.random.default_rng(42)
        index_x_in_rand = rng.integers(low=8, high=42, size=(10,)) + rng.normal(scale=0.3)
        index_y_in_rand = rng.integers(low=8, high=42, size=(10,)) + rng.normal(scale=0.3)
        points_x_rand, points_y_rand = raster.ij2xy(i=index_x_in_rand, j=index_y_in_rand, shift_area_or_point=shift_aop)

        for method in ["nearest", "linear"]:
            raster_points_mapcoords = raster.interp_points(
                (points_x_rand, points_y_rand),
                method=method,
                force_scipy_function="map_coordinates",
                shift_area_or_point=shift_aop,
                as_array=True,
            )
            raster_points_interpn = raster.interp_points(
                (points_x_rand, points_y_rand),
                method=method,
                force_scipy_function="interpn",
                shift_area_or_point=shift_aop,
                as_array=True,
            )

            # Not exactly equal in floating point precision since changes in Scipy 1.13.0,
            # see https://github.com/GlacioHack/geoutils/issues/533
            assert np.allclose(raster_points_mapcoords, raster_points_interpn)

        # Check that, outside the edge, the interpolation fails and returns a NaN
        index_x_edge_rand = [-0.5, -0.5, -0.5, 25, 25, 49.5, 49.5, 49.5]
        index_y_edge_rand = [-0.5, 25, 49.5, -0.5, 49.5, -0.5, 25, 49.5]

        points_x_rand, points_y_rand = raster.ij2xy(
            i=index_x_edge_rand, j=index_y_edge_rand, shift_area_or_point=shift_aop
        )

        # Test across all methods
        for method in ["nearest", "linear", "cubic", "quintic"]:
            raster_points_mapcoords_edge = raster.interp_points(
                (points_x_rand, points_y_rand),
                method=method,
                force_scipy_function="map_coordinates",
                shift_area_or_point=shift_aop,
                as_array=True,
            )
            raster_points_interpn_edge = raster.interp_points(
                (points_x_rand, points_y_rand),
                method=method,
                force_scipy_function="interpn",
                shift_area_or_point=shift_aop,
                as_array=True,
            )

            assert all(~np.isfinite(raster_points_mapcoords_edge))
            assert all(~np.isfinite(raster_points_interpn_edge))

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path])  # type: ignore
    @pytest.mark.parametrize(
        "method", ["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"]
    )  # type: ignore
    def test_interp_points__real(
        self, example: str, method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"]
    ) -> None:
        """
        Test interp_points for real data, checking in particular the propagation of nodata.

        For a random point (dimension 0) and a group of random points (dimension 1) in a real raster, we check the
        consistency of the output forcing to use either scipy.ndimage.map_coordinates or scipy.interpolate.interpn
        under-the-hood, or returning a regular-grid interpolator.
        """

        # Check the accuracy of the interpolation at an exact point, and between methods

        # Open and crop for speed
        r = gu.Raster(example)
        r = r.crop((r.bounds.left, r.bounds.bottom, r.bounds.left + r.res[0] * 50, r.bounds.bottom + r.res[1] * 50))
        r.set_area_or_point("Area", shift_area_or_point=False)

        # 1/ Test for an individual point (shape can be tricky in 1 dimension)
        itest = 10
        jtest = 10
        x, y = r.ij2xy(itest, jtest)
        val = r.interp_points((x, y), method=method, force_scipy_function="map_coordinates", as_array=True)[0]
        val_img = r.get_nanarray()[itest, jtest]
        # For a point exactly at a grid coordinate, only nearest and linear will match
        # (cubic modifies values at a grid coordinate)
        if method in ["nearest", "linear"]:
            assert val_img == pytest.approx(val, nan_ok=True)

        # Check the result is exactly the same for both methods
        val2 = r.interp_points((x, y), method=method, force_scipy_function="interpn", as_array=True)[0]
        assert val2 == pytest.approx(val, nan_ok=True)

        # Check that interp convert to latlon
        lat, lon = gu.projtools.reproject_to_latlon([x, y], in_crs=r.crs)
        val_latlon = r.interp_points((lat, lon), method=method, input_latlon=True, as_array=True)[0]
        assert val == pytest.approx(val_latlon, abs=0.0001, nan_ok=True)

        # 2/ Test for multiple points
        i = np.random.default_rng(42).integers(1, 49, size=10)
        j = np.random.default_rng(42).integers(1, 49, size=10)
        x, y = r.ij2xy(i, j)
        vals = r.interp_points((x, y), method=method, force_scipy_function="map_coordinates", as_array=True)
        vals2 = r.interp_points((x, y), method=method, force_scipy_function="interpn", as_array=True)

        assert np.array_equal(vals, vals2, equal_nan=True)

        # 3/ Test return_interpolator is consistent with above
        interp = _interp_points(
            r.get_nanarray(),
            transform=r.transform,
            area_or_point=r.area_or_point,
            points=(x, y),
            method=method,
            return_interpolator=True,
        )
        vals3 = interp((y, x))
        vals3 = np.array(np.atleast_1d(vals3), dtype=np.float32)

        assert np.array_equal(vals2, vals3, equal_nan=True)

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path])  # type: ignore
    @pytest.mark.parametrize(
        "method", ["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"]
    )  # type: ignore
    @pytest.mark.parametrize("dist", ["half_order_up", "half_order_down", 0, 1, 5])  # type: ignore
    def test_interp_point__nodata_propag(
        self,
        example: str,
        method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"],
        dist: Literal["half_order_up", "half_order_down"] | int,
    ) -> None:
        """
        Tests for nodata propagation of interp_points.

        We create artificial NaNs at certain pixels of real rasters (one with integer type, one floating type), then
        verify that interpolated values propagate these NaNs at the right distances for all methods, and across all
        nodata propagation distances.

        We also assess the precision of the interpolation under the default nodata propagation "half_order_up", by
        comparing interpolated values near nodata edges with and without the gap-filling of a constant placeholder
        value (= ensures the gap-filling of the array required for most methods that do not support NaN has little
        effect of interpolation).
        """
        # Open and crop for speed
        r = gu.Raster(example)
        r = r.crop((r.bounds.left, r.bounds.bottom, r.bounds.left + r.res[0] * 100, r.bounds.bottom + r.res[1] * 100))

        # 1/ Check the propagation of NaNs

        # 1.1/ Manual check for a specific point
        # Convert raster to float
        r = r.astype(np.float32)

        # Create a NaN at a given pixel (we know the landsat example has no NaNs to begin with)
        i0, j0 = (30, 30)  # This needs to be an area without NaN (and surroundings) in all test example
        r[i0, j0] = np.nan

        # Create a big NaN area in the middle (for more complex NaN propagation below)
        r[40:50, 40:50] = np.nan

        # All surrounding pixels with distance half the method order rounded up should be NaNs
        order = method_to_order[method]
        d = _get_dist_nodata_spread(order=order, dist_nodata_spread=dist)

        # Get indices of raster pixels within the right distance from NaNs
        indices_nan = [
            (i0 + i, j0 + j) for i in np.arange(-d, d + 1) for j in np.arange(-d, d + 1) if (np.abs(i) + np.abs(j)) <= d
        ]
        i, j = list(zip(*indices_nan))
        x, y = r.ij2xy(i, j)
        vals = r.interp_points(
            (x, y), method=method, force_scipy_function="map_coordinates", dist_nodata_spread=dist, as_array=True
        )
        vals2 = r.interp_points(
            (x, y), method=method, force_scipy_function="interpn", dist_nodata_spread=dist, as_array=True
        )

        assert all(np.isnan(np.atleast_1d(vals))) and all(np.isnan(np.atleast_1d(vals2)))

        # Same check for random coordinates within half a pixel of the above coordinates (falling exactly on grid
        # points)
        xoffset = np.random.default_rng(42).uniform(low=-0.5, high=0.5, size=len(x))
        yoffset = np.random.default_rng(42).uniform(low=-0.5, high=0.5, size=len(x))

        vals = r.interp_points(
            (x + xoffset, y + yoffset),
            method=method,
            force_scipy_function="map_coordinates",
            dist_nodata_spread=dist,
            as_array=True,
        )
        vals2 = r.interp_points(
            (x + xoffset, y + yoffset),
            method=method,
            force_scipy_function="interpn",
            dist_nodata_spread=dist,
            as_array=True,
        )

        assert all(np.isnan(np.atleast_1d(vals))) and all(np.isnan(np.atleast_1d(vals2)))

        # 1.2/ Check for all NaNs in the raster

        # We create the mask of dilated NaNs
        mask_nan = ~np.isfinite(r.get_nanarray())
        if d != 0:
            mask_nan_dilated = binary_dilation(mask_nan, iterations=d).astype("uint8")
        # (Zero iteration triggers a different behaviour than just "doing nothing" in binary_dilation, we override here)
        else:
            mask_nan_dilated = mask_nan.astype("uint8")
        # Get indices of the related pixels, convert to coordinates
        i, j = np.where(mask_nan_dilated)  # type: ignore
        x, y = r.ij2xy(i, j)
        # And interpolate at those coordinates
        vals = r.interp_points(
            (x, y), method=method, force_scipy_function="map_coordinates", dist_nodata_spread=dist, as_array=True
        )
        vals2 = r.interp_points(
            (x, y), method=method, force_scipy_function="interpn", dist_nodata_spread=dist, as_array=True
        )

        assert all(np.isnan(np.atleast_1d(vals))) and all(np.isnan(np.atleast_1d(vals2)))

        # 2/ Check that interpolated values at the edge of NaNs are valid + have small errors due to filling NaNs
        # with the nearest value during interpolation (thanks to the spreading of the nodata mask)

        # We compare values interpolated right at the edge of valid values near a NaN between
        # a/ Implementation of interp_points (that replaces NaNs by the nearest neighbour during interpolation)
        # b/ Raster filled with placeholder value then running interp_points

        # 2.1/ Manual check for a specific point

        # We get the indexes of valid pixels just at the edge of NaNs
        indices_edge = [
            (i0 + i, j0 + j)
            for i in np.arange(-d - 1, d + 2)
            for j in np.arange(-d - 1, d + 2)
            if (np.abs(i) + np.abs(j)) == d + 1
        ]
        i, j = list(zip(*indices_edge))
        x, y = r.ij2xy(i, j)
        # And get their interpolated value
        vals = r.interp_points(
            (x, y), method=method, force_scipy_function="map_coordinates", dist_nodata_spread=dist, as_array=True
        )
        vals2 = r.interp_points(
            (x, y), method=method, force_scipy_function="interpn", dist_nodata_spread=dist, as_array=True
        )

        # Then we fill the NaNs in the raster with a placeholder value of the raster mean
        r_arr = r.get_nanarray()
        r_arr[~np.isfinite(r_arr)] = np.nanmean(r_arr)
        r2 = r.copy(new_array=r_arr)

        # All raster values should be valid now
        assert np.all(np.isfinite(r_arr))

        # Only check the accuracy with the default NaN spreading (half-order rounded up), otherwise anything can happen
        if dist == "half_order_up":

            # Get the interpolated values
            vals_near = r2.interp_points(
                (x, y), method=method, force_scipy_function="map_coordinates", dist_nodata_spread=dist, as_array=True
            )
            vals2_near = r2.interp_points(
                (x, y), method=method, force_scipy_function="interpn", dist_nodata_spread=dist, as_array=True
            )

            # Both sets of values should be valid + within a relative tolerance of 0.01%
            assert np.allclose(vals, vals_near, equal_nan=False, rtol=10e-4)
            assert np.allclose(vals2, vals2_near, equal_nan=False, rtol=10e-4)

            # Same check for values within one pixel of the exact coordinates
            xoffset = np.random.default_rng(42).uniform(low=-0.5, high=0.5, size=len(x))
            yoffset = np.random.default_rng(42).uniform(low=-0.5, high=0.5, size=len(x))

            vals = r.interp_points(
                (x + xoffset, y + yoffset),
                method=method,
                force_scipy_function="map_coordinates",
                dist_nodata_spread=dist,
                as_array=True,
            )
            vals2 = r.interp_points(
                (x + xoffset, y + yoffset),
                method=method,
                force_scipy_function="interpn",
                dist_nodata_spread=dist,
                as_array=True,
            )
            vals_near = r2.interp_points(
                (x + xoffset, y + yoffset),
                method=method,
                force_scipy_function="map_coordinates",
                dist_nodata_spread=dist,
                as_array=True,
            )
            vals2_near = r2.interp_points(
                (x + xoffset, y + yoffset),
                method=method,
                force_scipy_function="interpn",
                dist_nodata_spread=dist,
                as_array=True,
            )

            # Both sets of values should be exactly the same, without any NaNs
            assert np.allclose(vals, vals_near, equal_nan=False, rtol=10e-4)
            assert np.allclose(vals2, vals2_near, equal_nan=False, rtol=10e-4)

            # 2.2/ Repeat the same for all edges of NaNs in the raster
            mask_dilated_plus_one = binary_dilation(mask_nan_dilated, iterations=1).astype(bool)
            mask_edge_dilated = np.logical_and(mask_dilated_plus_one, ~mask_nan_dilated.astype(bool))

            # Get indices of the related pixels, convert to coordinates
            i, j = np.where(mask_edge_dilated)  # type: ignore
            x, y = r.ij2xy(i, j)
            # And interpolate at those coordinates
            vals = r.interp_points(
                (x, y), method=method, force_scipy_function="map_coordinates", dist_nodata_spread=dist, as_array=True
            )
            vals2 = r.interp_points(
                (x, y), method=method, force_scipy_function="interpn", dist_nodata_spread=dist, as_array=True
            )
            vals_near = r2.interp_points(
                (x, y), method=method, force_scipy_function="map_coordinates", dist_nodata_spread=dist, as_array=True
            )
            vals2_near = r2.interp_points(
                (x, y), method=method, force_scipy_function="interpn", dist_nodata_spread=dist, as_array=True
            )

            # Both sets of values should be exactly the same, without any NaNs
            assert np.allclose(vals, vals_near, equal_nan=False, rtol=10e-4)
            assert np.allclose(vals2, vals2_near, equal_nan=False, rtol=10e-4)

    def test_reduce_points(self) -> None:
        """
        Test reduce points.
        """

        # -- Tests 1: Check based on indexed values --

        # Open raster
        r = gu.Raster(self.landsat_b4_crop_path)

        # A pixel center where all neighbouring coordinates are different:
        # array([[[237, 194, 239],
        #          [250, 173, 164],
        #          [255, 192, 128]]]
        itest0 = 19
        jtest0 = 21

        # Get coordinates at indices
        xtest0, ytest0 = r.ij2xy(itest0, jtest0, force_offset="center")

        # Check that the value at this coordinate is the same as when indexing
        z_val = r.reduce_points((xtest0, ytest0), as_array=True)
        z = r.data.data[itest0, jtest0]
        assert z == z_val

        # Check that the value is the same the other 4 corners of the pixel
        assert z == r.reduce_points((xtest0 + 0.49 * r.res[0], ytest0 - 0.49 * r.res[1]), as_array=True)
        assert z == r.reduce_points((xtest0 - 0.49 * r.res[0], ytest0 + 0.49 * r.res[1]), as_array=True)
        assert z == r.reduce_points((xtest0 - 0.49 * r.res[0], ytest0 - 0.49 * r.res[1]), as_array=True)
        assert z == r.reduce_points((xtest0 + 0.49 * r.res[0], ytest0 + 0.49 * r.res[1]), as_array=True)

        # -- Tests 2: check arguments work as intended --

        # 1/ Lat-lon argument check by getting the coordinates of our last test point
        lat, lon = reproject_to_latlon(points=[xtest0, ytest0], in_crs=r.crs)
        z_val_2 = r.reduce_points((lon, lat), input_latlon=True, as_array=True)
        assert z_val == z_val_2

        # 2/ Band argument
        # Get the band indexes for the multi-band Raster
        r_multi = gu.Raster(self.landsat_rgb_path)
        itest, jtest = r_multi.xy2ij(xtest0, ytest0)
        itest = int(itest[0])
        jtest = int(jtest[0])
        # Extract the values
        z_band1 = r_multi.reduce_points((xtest0, ytest0), band=1, as_array=True)
        z_band2 = r_multi.reduce_points((xtest0, ytest0), band=2, as_array=True)
        z_band3 = r_multi.reduce_points((xtest0, ytest0), band=3, as_array=True)
        # Compare to the Raster array slice
        assert list(r_multi.data[:, itest, jtest]) == [z_band1, z_band2, z_band3]

        # 3/ Masked argument
        r_multi.data[:, itest, jtest] = np.ma.masked
        z_not_ma = r_multi.reduce_points((xtest0, ytest0), band=1, as_array=True)
        assert not np.ma.is_masked(z_not_ma)
        z_ma = r_multi.reduce_points((xtest0, ytest0), band=1, masked=True, as_array=True)
        assert np.ma.is_masked(z_ma)

        # 4/ Window argument
        val_window, z_window = r_multi.reduce_points(
            (xtest0, ytest0), band=1, window=3, masked=True, return_window=True, as_array=True
        )
        assert (
            val_window
            == np.ma.mean(r_multi.data[0, itest - 1 : itest + 2, jtest - 1 : jtest + 2])
            == np.ma.mean(z_window)
        )
        assert np.array_equal(z_window, r_multi.data[0, itest - 1 : itest + 2, jtest - 1 : jtest + 2])

        # 5/ Reducer function argument
        val_window2 = r_multi.reduce_points(
            (xtest0, ytest0), band=1, window=3, masked=True, reducer_function=np.ma.median, as_array=True
        )
        assert val_window2 == np.ma.median(r_multi.data[0, itest - 1 : itest + 2, jtest - 1 : jtest + 2])

        # -- Tests 3: check that errors are raised when supposed for non-boolean arguments --

        # Verify that passing a window that is not a whole number fails
        with pytest.raises(ValueError, match=re.escape("Window must be a whole number.")):
            r.reduce_points((xtest0, ytest0), window=3.5)  # type: ignore
        # Same for an odd number
        with pytest.raises(ValueError, match=re.escape("Window must be an odd number.")):
            r.reduce_points((xtest0, ytest0), window=4)
        # But a window that is a whole number as a float works
        r.reduce_points((xtest0, ytest0), window=3.0)  # type: ignore

        # -- Tests 4: check that passing an array-like object works

        # For simple coordinates
        x_coords = [xtest0, xtest0 + 10]
        y_coords = [ytest0, ytest0 - 10]
        vals = r_multi.reduce_points((x_coords, y_coords), as_array=True)
        val0, win0 = r_multi.reduce_points((x_coords[0], y_coords[0]), return_window=True, as_array=True)
        val1, win1 = r_multi.reduce_points((x_coords[1], y_coords[1]), return_window=True, as_array=True)

        assert len(vals) == len(x_coords)
        assert np.array_equal(vals[0], val0, equal_nan=True)
        assert np.array_equal(vals[1], val1, equal_nan=True)

        # With a return window argument
        vals, windows = r_multi.reduce_points((x_coords, y_coords), return_window=True, as_array=True)
        assert len(windows) == len(x_coords)
        assert np.array_equal(windows[0], win0, equal_nan=True)
        assert np.array_equal(windows[1], win1, equal_nan=True)

        # -- Tests 5 -- Check image corners and latlon argument

        # Lower right pixel
        x, y = [r.bounds.right - r.res[0] / 2, r.bounds.bottom + r.res[1] / 2]
        lat, lon = reproject_to_latlon([x, y], r.crs)
        assert (
            r.reduce_points((x, y), as_array=True)
            == r.reduce_points((lon, lat), input_latlon=True, as_array=True)
            == r.data[-1, -1]
        )

        # One pixel above
        x, y = [r.bounds.right - r.res[0] / 2, r.bounds.bottom + 3 * r.res[1] / 2]
        lat, lon = reproject_to_latlon([x, y], r.crs)
        assert (
            r.reduce_points((x, y), as_array=True)
            == r.reduce_points((lon, lat), input_latlon=True, as_array=True)
            == r.data[-2, -1]
        )

        # One pixel left
        x, y = [r.bounds.right - 3 * r.res[0] / 2, r.bounds.bottom + r.res[1] / 2]
        lat, lon = reproject_to_latlon([x, y], r.crs)
        assert (
            r.reduce_points((x, y), as_array=True)
            == r.reduce_points((lon, lat), input_latlon=True, as_array=True)
            == r.data[-1, -2]
        )



"""
Interpolate raster at points
============================

This example demonstrates the 2D interpolation of raster values to points using :func:`~geoutils.Raster.interp_points`.
"""

# %%
# We open an example raster, a digital elevation model in South America.

# sphinx_gallery_thumbnail_number = 2
import geoutils as gu

filename_rast = gu.examples.get_path("exploradores_aster_dem")
rast = gu.Raster(filename_rast)
rast = rast.crop([rast.bounds.left, rast.bounds.bottom, rast.bounds.left + 2000, rast.bounds.bottom + 2000])

# Plot the raster
rast.plot(cmap="terrain")

# %%
# We generate a random subsample of 100 coordinates to interpolate.

import numpy as np

rng = np.random.default_rng(42)
x_coords = rng.uniform(rast.bounds.left + 50, rast.bounds.right - 50, 50)
y_coords = rng.uniform(rast.bounds.bottom + 50, rast.bounds.top - 50, 50)

pc = rast.interp_points(points=(x_coords, y_coords))

# %%
# We plot the resulting point cloud
pc.plot(ax="new", cmap="terrain", marker="x", cbar_title="Elevation (m)")

# %%
# .. important::
#       The interpretation of where raster values are located can differ. The parameter ``shift_area_or_point`` (off by default) can be turned on to ensure
#       that the pixel interpretation of your dataset is correct.

# %%
# Let's look and redefine our pixel interpretation into ``"Point"``. This will shift interpolation by half a pixel.

rast.area_or_point
rast.area_or_point = "Point"

# %%
# We can interpolate again by shifting according to our interpretation, and changing the resampling algorithm (default to "linear").

pc_shifted = rast.interp_points(points=(x_coords, y_coords), shift_area_or_point=True, method="quintic")
np.nanmean(pc - pc_shifted)

# %%
# The mean difference in interpolated values is quite significant, with a 2-meter bias!



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functionalities for geotransformations of raster objects.
"""

from __future__ import annotations

import warnings
from typing import Iterable, Literal

import affine
import numpy as np
import rasterio as rio
from rasterio.crs import CRS
from rasterio.enums import Resampling

import geoutils as gu
from geoutils import profiler
from geoutils._typing import DTypeLike, MArrayNum
from geoutils.raster._geotransformations import (
    _get_reproj_params,
    _is_reproj_needed,
    _rio_reproject,
    _user_input_reproject,
)
from geoutils.raster.distributed_computing.multiproc import _multiproc_reproject
from geoutils.raster.georeferencing import _cast_pixel_interpretation

##############
# 1/ REPROJECT
##############


@profiler.profile("geoutils.raster.geotransformations._reproject", memprof=True)  # type: ignore
def _reproject(
    source_raster: gu.Raster,
    ref: gu.Raster,
    crs: CRS | str | int | None = None,
    res: float | Iterable[float] | None = None,
    grid_size: tuple[int, int] | None = None,
    bounds: dict[str, float] | rio.coords.BoundingBox | None = None,
    nodata: int | float | None = None,
    dtype: DTypeLike | None = None,
    resampling: Resampling | str = Resampling.bilinear,
    force_source_nodata: int | float | None = None,
    silent: bool = False,
    n_threads: int = 0,
    memory_limit: int = 64,
    multiproc_config: gu.raster.MultiprocConfig | None = None,
) -> tuple[bool, MArrayNum | None, affine.Affine | None, CRS | None, int | float | None]:
    """
    Reproject raster. See Raster.reproject() for details.
    """

    # 1/ Process user input
    crs, dtype, src_nodata, nodata, res, bounds = _user_input_reproject(
        source_raster=source_raster,
        ref=ref,
        crs=crs,
        bounds=bounds,
        res=res,
        nodata=nodata,
        dtype=dtype,
        force_source_nodata=force_source_nodata,
    )

    # 2/ Derive georeferencing parameters for reprojection (transform, grid size)
    reproj_kwargs = _get_reproj_params(
        source_raster=source_raster,
        crs=crs,
        res=res,
        grid_size=grid_size,
        bounds=bounds,
        dtype=dtype,
        src_nodata=src_nodata,
        nodata=nodata,
        resampling=resampling,
    )

    # 3/ Check if reprojection is needed, otherwise return source raster with warning
    if _is_reproj_needed(src_shape=source_raster.shape, reproj_kwargs=reproj_kwargs):
        if (nodata == src_nodata) or (nodata is None):
            if not silent:
                warnings.warn("Output projection, bounds and grid size are identical -> returning self (not a copy!)")
            return True, None, None, None, None

        elif nodata is not None:
            if not silent:
                warnings.warn(
                    "Only nodata is different, consider using the 'set_nodata()' method instead'\
                ' -> returning self (not a copy!)"
                )
            return True, None, None, None, None

    # 4/ Check reprojection is possible (boolean raster will be converted, so no need to check)
    if np.dtype(source_raster.dtype) != bool and (src_nodata is None and np.sum(source_raster.data.mask) > 0):
        raise ValueError(
            "No nodata set, set one for the raster with self.set_nodata() or use a temporary one "
            "with `force_source_nodata`."
        )

    # 5/ Perform reprojection
    reproj_kwargs.update({"num_threads": n_threads, "warp_mem_limit": memory_limit})
    if multiproc_config is not None:
        _multiproc_reproject(source_raster, config=multiproc_config, **reproj_kwargs)
        return False, None, None, None, None

    else:
        # All masked values must be set to a nodata value for rasterio's reproject to work properly
        src_arr = source_raster.data.data
        src_mask = np.ma.getmaskarray(source_raster.data)
        dst_arr, dst_mask = _rio_reproject(src_arr=src_arr, src_mask=src_mask, reproj_kwargs=reproj_kwargs)
        # Set mask
        dst_arr = np.ma.masked_array(data=dst_arr, mask=dst_mask, fill_value=nodata)

        return False, dst_arr, reproj_kwargs["dst_transform"], reproj_kwargs["dst_crs"], reproj_kwargs["dst_nodata"]


#########
# 2/ CROP
#########


@profiler.profile("geoutils.raster.geotransformations._crop", memprof=True)  # type: ignore
def _crop(
    source_raster: gu.Raster,
    bbox: gu.Raster | gu.Vector | list[float] | tuple[float, ...],
    mode: Literal["match_pixel"] | Literal["match_extent"] = "match_pixel",
    distance_unit: Literal["georeferenced", "pixel"] = "georeferenced",
) -> tuple[MArrayNum, affine.Affine]:
    """Crop raster. See details in Raster.crop()."""

    assert mode in ["match_extent", "match_pixel"], "mode must be one of 'match_pixel', 'match_extent'"
    assert distance_unit in ["georeferenced", "pixel"], "distance_unit must be 'georeferenced' or 'pixel'"

    if isinstance(bbox, (gu.Raster, gu.Vector)):
        # For another Vector or Raster, we reproject the bounding box in the same CRS as self
        xmin, ymin, xmax, ymax = bbox.get_bounds_projected(out_crs=source_raster.crs)
        if isinstance(bbox, gu.Raster):
            # Raise a warning if the reference is a raster that has a different pixel interpretation
            _cast_pixel_interpretation(source_raster.area_or_point, bbox.area_or_point)
    elif isinstance(bbox, (list, tuple)):
        if distance_unit == "georeferenced":
            xmin, ymin, xmax, ymax = bbox
        else:
            colmin, rowmin, colmax, rowmax = bbox
            xmin, ymax = rio.transform.xy(source_raster.transform, rowmin, colmin, offset="ul")
            xmax, ymin = rio.transform.xy(source_raster.transform, rowmax, colmax, offset="ul")
    else:
        raise ValueError("cropGeom must be a Raster, Vector, or list of coordinates.")

    if mode == "match_pixel":
        # Finding the intersection of requested bounds and original bounds, cropped to image shape
        ref_win = rio.windows.from_bounds(xmin, ymin, xmax, ymax, transform=source_raster.transform)
        self_win = rio.windows.from_bounds(*source_raster.bounds, transform=source_raster.transform).crop(
            *source_raster.shape
        )
        final_window = ref_win.intersection(self_win).round_lengths().round_offsets()

        # Update bounds and transform accordingly
        new_xmin, new_ymin, new_xmax, new_ymax = rio.windows.bounds(final_window, transform=source_raster.transform)
        tfm = rio.transform.from_origin(new_xmin, new_ymax, *source_raster.res)

        if source_raster.is_loaded:
            # In case data is loaded on disk, can extract directly from np array
            (rowmin, rowmax), (colmin, colmax) = final_window.toranges()
            crop_img = source_raster.data[..., rowmin:rowmax, colmin:colmax]

        else:

            assert source_raster._disk_shape is not None  # This should not be the case, sanity check to make mypy happy

            # If data was not loaded, and self's transform was updated (e.g. due to downsampling) need to
            # get the Window corresponding to on disk data
            ref_win_disk = rio.windows.from_bounds(
                new_xmin, new_ymin, new_xmax, new_ymax, transform=source_raster._disk_transform
            )
            self_win_disk = rio.windows.from_bounds(
                *source_raster.bounds, transform=source_raster._disk_transform
            ).crop(*source_raster._disk_shape[1:])
            final_window_disk = ref_win_disk.intersection(self_win_disk).round_lengths().round_offsets()

            # Round up to downsampling size, to match __init__
            final_window_disk = rio.windows.round_window_to_full_blocks(
                final_window_disk, ((source_raster._downsample, source_raster._downsample),)
            )

            # Load data for "on_disk" window but out_shape matching in-memory transform -> enforce downsampling
            # AD (24/04/24): Note that the same issue as #447 occurs here when final_window_disk extends beyond
            # self's bounds. Using option `boundless=True` solves the issue but causes other tests to fail
            # This should be fixed with #447 and previous line would be obsolete.
            with rio.open(source_raster.filename) as raster:
                crop_img = raster.read(
                    indexes=source_raster._bands,
                    masked=source_raster._masked,
                    window=final_window_disk,
                    out_shape=(final_window.height, final_window.width),
                )

            # Squeeze first axis for single-band
            if crop_img.ndim == 3 and crop_img.shape[0] == 1:
                crop_img = crop_img.squeeze(axis=0)

    else:
        bbox = rio.coords.BoundingBox(left=xmin, bottom=ymin, right=xmax, top=ymax)
        out_rst = source_raster.reproject(bounds=bbox)  # should we instead raise an issue and point to reproject?
        crop_img = out_rst.data
        tfm = out_rst.transform

    return crop_img, tfm


##############
# 3/ TRANSLATE
##############


@profiler.profile("geoutils.raster.geotransformations._translate", memprof=True)  # type: ignore
def _translate(
    transform: affine.Affine,
    xoff: float,
    yoff: float,
    distance_unit: Literal["georeferenced", "pixel"] = "georeferenced",
) -> affine.Affine:
    """
    Translate geotransform horizontally, either in pixels or georeferenced units.

    :param transform: Input geotransform.
    :param xoff: Translation x offset.
    :param yoff: Translation y offset.
    :param distance_unit: Distance unit, either 'georeferenced' (default) or 'pixel'.

    :return: Translated transform.
    """

    if distance_unit not in ["georeferenced", "pixel"]:
        raise ValueError("Argument 'distance_unit' should be either 'pixel' or 'georeferenced'.")

    # Get transform
    dx, b, xmin, d, dy, ymax = list(transform)[:6]

    # Convert pixel offsets to georeferenced units
    if distance_unit == "pixel":
        xoff *= dx
        yoff *= dy

    return rio.transform.Affine(dx, b, xmin + xoff, d, dy, ymax + yoff)



from __future__ import annotations

import numpy as np
import pytest

import geoutils as gu
from geoutils import examples


class TestGeoreferencing:

    landsat_b4_path = examples.get_path_test("everest_landsat_b4")
    aster_dem_path = examples.get_path_test("exploradores_aster_dem")
    landsat_rgb_path = examples.get_path_test("everest_landsat_rgb")
    landsat_b4_crop_path = examples.get_path_test("everest_landsat_b4_cropped")

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path, landsat_rgb_path])  # type: ignore
    def test_ij2xy(self, example: str) -> None:
        """Test the outputs of ij2xy and that the function is reversible with xy2ij."""

        # Open raster
        rst = gu.Raster(example)
        xmin, ymin, xmax, ymax = rst.bounds

        # Check ij2xy manually for the four corners and center
        # With "force_offset", no considerations of pixel interpolation

        # With offset="center", should be pixel center
        xmin_center = xmin + rst.res[0] / 2
        ymin_center = ymin + rst.res[1] / 2
        xmax_center = xmax - rst.res[0] / 2
        ymax_center = ymax - rst.res[1] / 2
        assert rst.ij2xy([0], [0], force_offset="center") == ([xmin_center], [ymax_center])
        assert rst.ij2xy([rst.shape[0] - 1], [0], force_offset="center") == ([xmin_center], [ymin_center])
        assert rst.ij2xy([0], [rst.shape[1] - 1], force_offset="center") == ([xmax_center], [ymax_center])
        assert rst.ij2xy([rst.shape[0] - 1], [rst.shape[1] - 1], force_offset="center") == (
            [xmax_center],
            [ymin_center],
        )

        # Same checks for offset="ll", "ul", "ur", "lr
        lims_ll = [xmin, ymin, xmax - rst.res[0], ymax - rst.res[1]]
        lims_ul = [xmin, ymin + rst.res[1], xmax - rst.res[0], ymax]
        lims_lr = [xmin + rst.res[0], ymin, xmax, ymax - rst.res[1]]
        lims_ur = [xmin + rst.res[0], ymin + rst.res[1], xmax, ymax]
        offsets = ["ll", "ul", "lr", "ur"]
        list_lims = [lims_ll, lims_ul, lims_lr, lims_ur]
        for i in range(len(list_lims)):
            offset = offsets[i]
            lim = list_lims[i]
            assert rst.ij2xy([0], [0], force_offset=offset) == ([lim[0]], [lim[3]])
            assert rst.ij2xy([rst.shape[0] - 1], [0], force_offset=offset) == ([lim[0]], [lim[1]])
            assert rst.ij2xy([0], [rst.shape[1] - 1], force_offset=offset) == ([lim[2]], [lim[3]])
            assert rst.ij2xy([rst.shape[0] - 1], [rst.shape[1] - 1], force_offset=offset) == ([lim[2]], [lim[1]])

        # Check that default coordinate is upper-left

        # With shift from pixel interpretation, coordinates will be half a pixel back for "Point"
        if rst.area_or_point is not None and rst.area_or_point == "Point" and gu.config["shift_area_or_point"]:
            # Shift is backward in X, forward in Y
            lims_ul[0] = lims_ul[0] - 0.5 * rst.res[0]
            lims_ul[1] = lims_ul[1] + 0.5 * rst.res[1]
            lims_ul[2] = lims_ul[2] - 0.5 * rst.res[0]
            lims_ul[3] = lims_ul[3] + 0.5 * rst.res[1]

        assert rst.ij2xy([0], [0]) == ([lims_ul[0]], [lims_ul[3]])
        assert rst.ij2xy([rst.shape[0] - 1], [0]) == ([lims_ul[0]], [lims_ul[1]])
        assert rst.ij2xy([0], [rst.shape[1] - 1]) == ([lims_ul[2]], [lims_ul[3]])
        assert rst.ij2xy([rst.shape[0] - 1], [rst.shape[1] - 1]) == ([lims_ul[2]], [lims_ul[1]])

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path, landsat_rgb_path])  # type: ignore
    def test_xy2ij_ij2xy_reversible(self, example: str) -> None:

        # Open raster
        rst = gu.Raster(example)
        xmin, ymin, xmax, ymax = rst.bounds

        # We generate random points within the boundaries of the image
        rng = np.random.default_rng(42)
        xrand = rng.integers(low=0, high=rst.width, size=(10,)) * list(rst.transform)[0] + xmin
        yrand = ymax + rng.integers(low=0, high=rst.height, size=(10,)) * list(rst.transform)[4]

        # Test reversibility for any point or area interpretation
        i, j = rst.xy2ij(xrand, yrand)
        xnew, ynew = rst.ij2xy(i, j)
        assert all(xnew == xrand)
        assert all(ynew == yrand)

        # TODO: clarify this weird behaviour of rasterio.index with floats?
        # r.ds.index(x, y)
        # Out[33]: (75, 301)
        # r.ds.index(x, y, op=np.float32)
        # Out[34]: (75.0, 302.0)

    def test_xy2ij(self) -> None:
        """Test xy2ij with shift_area_or_point argument, and compare to interp_points function for consistency."""

        # First, we try on a Raster with a Point interpretation: values interpolated at the center of pixel
        r = gu.Raster(self.landsat_b4_path)
        r.set_area_or_point("Point")
        assert r.area_or_point == "Point"
        xmin, ymin, xmax, ymax = r.bounds

        # We generate random points within the boundaries of the image
        rng = np.random.default_rng(42)
        xrand = rng.integers(low=0, high=r.width, size=(10,)) * list(r.transform)[0] + xmin
        yrand = ymax + rng.integers(low=0, high=r.height, size=(10,)) * list(r.transform)[4]

        # Get decimal indexes based on "Point", should refer to the corner still (shift False by default)
        i, j = r.xy2ij(xrand, yrand, shift_area_or_point=False)
        assert np.all(i % 1 == 0)
        assert np.all(j % 1 == 0)

        # Those should all be .5 because values refer to the center and are shifted
        i, j = r.xy2ij(xrand, yrand, shift_area_or_point=True)
        assert np.all(i % 1 == 0.5)
        assert np.all(j % 1 == 0.5)

        # Force "Area", should refer to corner
        r.set_area_or_point("Area", shift_area_or_point=False)
        i, j = r.xy2ij(xrand, yrand, shift_area_or_point=True)
        assert np.all(i % 1 == 0)
        assert np.all(j % 1 == 0)

        # Now, we calculate the mean of values in each 2x2 slices of the data, and compare with interpolation at order 1
        list_z_ind = []
        img = r.data
        for k in range(len(xrand)):
            # 2x2 slices
            z_ind = np.mean(
                img[
                    slice(int(np.floor(i[k])), int(np.ceil(i[k])) + 1),
                    slice(int(np.floor(j[k])), int(np.ceil(j[k])) + 1),
                ]
            )
            list_z_ind.append(z_ind)

        # First order interpolation
        rpts = r.interp_points((xrand, yrand), method="linear", as_array=True)
        # The values interpolated should be equal
        assert np.array_equal(np.array(list_z_ind, dtype=np.float32), rpts, equal_nan=True)

        # Test there is no failure with random coordinates (edge effects, etc)
        xrand = rng.uniform(low=xmin, high=xmax, size=(1000,))
        yrand = rng.uniform(low=ymin, high=ymax, size=(1000,))
        r.interp_points((xrand, yrand), as_array=True)

        # Second, test after a crop: the Raster now has an Area interpretation, those should fall right on the integer
        # pixel indexes
        r2 = gu.Raster(self.landsat_b4_crop_path)
        r.crop(r2)
        assert r.area_or_point == "Area"

        xmin, ymin, xmax, ymax = r.bounds

        # We can test with several method for the exact indexes: interp, and simple read should
        # give back the same values that fall right on the coordinates
        xrand = rng.integers(low=0, high=r.width, size=(10,)) * list(r.transform)[0] + xmin
        yrand = ymax + rng.integers(low=0, high=r.height, size=(10,)) * list(r.transform)[4]
        # By default, i and j are returned as integers
        i, j = r.xy2ij(xrand, yrand, op=np.float32)
        list_z_ind = []
        img = r.data
        for k in range(len(xrand)):
            # We directly sample the values
            z_ind = img[int(i[k]), int(j[k])]
            list_z_ind.append(z_ind)

        rpts = r.interp_points((xrand, yrand), method="linear", as_array=True)

        assert np.array_equal(np.array(list_z_ind, dtype=np.float32), rpts, equal_nan=True)

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path, landsat_rgb_path])  # type: ignore
    def test_coords(self, example: str) -> None:

        img = gu.Raster(example)

        # With lower left argument
        xx0, yy0 = img.coords(grid=False, force_offset="ll")

        assert len(xx0) == img.width
        assert len(yy0) == img.height

        assert xx0[0] == pytest.approx(img.bounds.left)
        assert xx0[-1] == pytest.approx(img.bounds.right - img.res[0])
        if img.res[1] > 0:
            assert yy0[0] == pytest.approx(img.bounds.bottom)
            assert yy0[-1] == pytest.approx(img.bounds.top - img.res[1])
        else:
            # Currently not covered by test image
            assert yy0[0] == pytest.approx(img.bounds.top)
            assert yy0[-1] == pytest.approx(img.bounds.bottom + img.res[1])

        # With center argument
        xx, yy = img.coords(grid=False, force_offset="center")
        hx = img.res[0] / 2
        hy = img.res[1] / 2
        assert xx[0] == pytest.approx(img.bounds.left + hx)
        assert xx[-1] == pytest.approx(img.bounds.right - hx)
        if img.res[1] > 0:
            assert yy[0] == pytest.approx(img.bounds.bottom + hy)
            assert yy[-1] == pytest.approx(img.bounds.top - hy)
        else:
            # Currently not covered by test image
            assert yy[-1] == pytest.approx(img.bounds.top + hy)
            assert yy[0] == pytest.approx(img.bounds.bottom - hy)

        # With grid argument (default argument, repeated here for clarity)
        xxgrid, yygrid = img.coords(grid=True, force_offset="ll")
        assert np.array_equal(xxgrid, np.repeat(xx0[np.newaxis, :], img.height, axis=0))
        assert np.array_equal(yygrid, np.flipud(np.repeat(yy0[:, np.newaxis], img.width, axis=1)))



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionalities at the interface of rasters and point clouds."""

from __future__ import annotations

from typing import Iterable, Literal

import affine
import geopandas as gpd
import numpy as np
import rasterio as rio
from rasterio.crs import CRS

import geoutils as gu
from geoutils._typing import NDArrayNum
from geoutils.raster.array import get_mask_from_array
from geoutils.raster.georeferencing import _default_nodata, _xy2ij
from geoutils.stats import subsample_array


def _regular_pointcloud_to_raster(
    pointcloud: gpd.GeoDataFrame | gu.PointCloud,
    grid_coords: tuple[NDArrayNum, NDArrayNum] = None,
    transform: rio.transform.Affine = None,
    shape: tuple[int, int] = None,
    nodata: int | float | None = None,
    data_column_name: str = "b1",
    area_or_point: Literal["Area", "Point"] = "Point",
) -> tuple[NDArrayNum, affine.Affine, CRS, int | float | None, Literal["Area", "Point"]]:
    """
    Convert a regular point cloud to a raster. See Raster.from_pointcloud_regular() for details.
    """

    # Extract geodataframe and data column name depending on input
    if isinstance(pointcloud, gu.PointCloud):
        gdf_pc = pointcloud.ds
        data_column_name = pointcloud.data_column
    else:
        gdf_pc = pointcloud
        data_column_name = data_column_name

    # Get transform and shape from input
    if grid_coords is not None:

        # Input checks
        if (
            not isinstance(grid_coords, tuple)
            or not (isinstance(grid_coords[0], np.ndarray) and grid_coords[0].ndim == 1)
            or not (isinstance(grid_coords[1], np.ndarray) and grid_coords[1].ndim == 1)
        ):
            raise TypeError("Input grid coordinates must be 1D arrays.")

        diff_x = np.diff(grid_coords[0])
        diff_y = np.diff(grid_coords[1])

        if not all(diff_x == diff_x[0]) and all(diff_y == diff_y[0]):
            raise ValueError("Grid coordinates must be regular (equally spaced, independently along X and Y).")

        # Build transform from min X, max Y and step in both
        out_transform = rio.transform.from_origin(np.min(grid_coords[0]), np.max(grid_coords[1]), diff_x[0], diff_y[0])
        # Y is first axis, X is second axis
        out_shape = (len(grid_coords[1]), len(grid_coords[0]))

    elif transform is not None and shape is not None:

        out_transform = transform
        out_shape = shape

    else:
        raise ValueError("Either grid coordinates or both geotransform and shape must be provided.")

    # Create raster from inputs, with placeholder data for now
    dtype = gdf_pc[data_column_name].dtype
    out_nodata = nodata if nodata is not None else _default_nodata(dtype)
    arr = np.ones(out_shape, dtype=dtype)

    # Get indexes of point cloud coordinates in the raster, forcing no shift
    i, j = _xy2ij(
        x=gdf_pc.geometry.x.values,
        y=gdf_pc.geometry.y.values,
        shift_area_or_point=False,
        transform=out_transform,
        area_or_point=area_or_point,
    )

    # If coordinates are not integer type (forced in xy2ij), then some points are not falling on exact coordinates
    if not np.issubdtype(i.dtype, np.integer) or not np.issubdtype(i.dtype, np.integer):
        raise ValueError("Some point cloud coordinates differ from the grid coordinates.")

    # Set values
    mask = np.ones(np.shape(arr), dtype=bool)
    mask[i, j] = False
    arr[i, j] = gdf_pc[data_column_name].values

    # Set output values
    raster_arr = np.ma.masked_array(data=arr, mask=mask)

    return raster_arr, out_transform, gdf_pc.crs, out_nodata, area_or_point


def _raster_to_pointcloud(
    source_raster: gu.Raster,
    data_column_name: str = "b1",
    data_band: int = 1,
    auxiliary_data_bands: list[int] | None = None,
    auxiliary_column_names: list[str] | None = None,
    subsample: float | int = 1,
    skip_nodata: bool = True,
    as_array: bool = False,
    random_state: int | np.random.Generator | None = None,
    force_pixel_offset: Literal["center", "ul", "ur", "ll", "lr"] = "ul",
) -> NDArrayNum | gu.PointCloud:
    """
    Convert a raster to a point cloud. See Raster.to_pointcloud() for details.
    """

    # Input checks

    # Main data column checks
    if not isinstance(data_column_name, str):
        raise ValueError("Data column name must be a string.")
    if not (isinstance(data_band, int) and data_band >= 1 and data_band <= source_raster.count):
        raise ValueError(
            f"Data band number must be an integer between 1 and the total number of bands ({source_raster.count})."
        )

    # Rename data column if a different band is selected but the name is still default
    if data_band != 1 and data_column_name == "b1":
        data_column_name = "b" + str(data_band)

    # Auxiliary data columns checks
    if auxiliary_column_names is not None and auxiliary_data_bands is None:
        raise ValueError("Passing auxiliary column names requires passing auxiliary data band numbers as well.")
    if auxiliary_data_bands is not None:
        if not (isinstance(auxiliary_data_bands, Iterable) and all(isinstance(b, int) for b in auxiliary_data_bands)):
            raise ValueError("Auxiliary data band number must be an iterable containing only integers.")
        if any((1 > b or source_raster.count < b) for b in auxiliary_data_bands):
            raise ValueError(
                f"Auxiliary data band numbers must be between 1 and the total number of bands ({source_raster.count})."
            )
        if data_band in auxiliary_data_bands:
            raise ValueError(
                f"Main data band {data_band} should not be listed in auxiliary data bands {auxiliary_data_bands}."
            )

        # Ensure auxiliary column name is defined if auxiliary data bands is not None
        if auxiliary_column_names is not None:
            if not (
                isinstance(auxiliary_column_names, Iterable) and all(isinstance(b, str) for b in auxiliary_column_names)
            ):
                raise ValueError("Auxiliary column names must be an iterable containing only strings.")
            if not len(auxiliary_column_names) == len(auxiliary_data_bands):
                raise ValueError(
                    f"Length of auxiliary column name and data band numbers should be the same, "
                    f"found {len(auxiliary_column_names)} and {len(auxiliary_data_bands)} respectively."
                )

        else:
            auxiliary_column_names = [f"b{i}" for i in auxiliary_data_bands]

        # Define bigger list with all bands and names
        all_bands = [data_band] + auxiliary_data_bands
        all_column_names = [data_column_name] + auxiliary_column_names

    else:
        all_bands = [data_band]
        all_column_names = [data_column_name]

    # If subsample is the entire array, load it to optimize speed
    if subsample == 1 and not source_raster.is_loaded:
        source_raster.load(bands=all_bands)

    # Band indexes in the array are band number minus one
    all_indexes = [b - 1 for b in all_bands]

    # We do 2D subsampling on the data band only, regardless of valid masks on other bands
    if skip_nodata:
        if source_raster.is_loaded:
            if source_raster.count == 1:
                self_mask = get_mask_from_array(
                    source_raster.data
                )  # This is to avoid the case where the mask is just "False"
            else:
                self_mask = get_mask_from_array(
                    source_raster.data[data_band - 1, :, :]
                )  # This is to avoid the case where the mask is just "False"
            valid_mask = ~self_mask

        # Load only mask of valid data from disk if array not loaded
        else:
            valid_mask = ~source_raster._load_only_mask(bands=data_band)
    # If we are not skipping nodata values, valid mask is everywhere
    else:
        if source_raster.count == 1:
            valid_mask = np.ones(source_raster.data.shape, dtype=bool)
        else:
            valid_mask = np.ones(source_raster.data[0, :].shape, dtype=bool)

    # Get subsample on valid mask
    # Build a low memory boolean masked array with invalid values masked to pass to subsampling
    ma_valid = np.ma.masked_array(data=np.ones(np.shape(valid_mask), dtype=bool), mask=~valid_mask)
    # Take a subsample within the valid values
    indices = subsample_array(array=ma_valid, subsample=subsample, random_state=random_state, return_indices=True)

    # If the Raster is loaded, pick from the data while ignoring the mask
    if source_raster.is_loaded:
        if source_raster.count == 1:
            pixel_data = source_raster.data[indices[0], indices[1]]
        else:
            # TODO: Combining both indexes at once could reduce memory usage?
            pixel_data = source_raster.data[all_indexes, :][:, indices[0], indices[1]]

    # Otherwise use rasterio.sample to load only requested pixels
    else:
        # Extract the coordinates at subsampled pixels with valid data
        # To extract data, we always use "upper left" which rasterio interprets as the exact raster coordinates
        # Further below we redefine output coordinates based on point interpretation
        x_coords, y_coords = (np.array(a) for a in source_raster.ij2xy(indices[0], indices[1], force_offset="ul"))

        with rio.open(source_raster.filename) as raster:
            # Rasterio uses indexes (starts at 1)
            pixel_data = np.array(list(raster.sample(zip(x_coords, y_coords), indexes=all_bands))).T

    # At this point there should not be any nodata anymore, so we can transform everything to normal array
    if np.ma.isMaskedArray(pixel_data):
        pixel_data = pixel_data.data

    # If nodata values were not skipped, convert them to NaNs and change data type
    if skip_nodata is False:
        pixel_data = pixel_data.astype("float32")
        pixel_data[pixel_data == source_raster.nodata] = np.nan

    # Now we force the coordinates we define for the point cloud, according to pixel interpretation
    x_coords_2, y_coords_2 = (
        np.array(a) for a in source_raster.ij2xy(indices[0], indices[1], force_offset=force_pixel_offset)
    )

    if not as_array:
        pc = gpd.GeoDataFrame(
            pixel_data.T,
            columns=all_column_names,
            geometry=gpd.points_from_xy(x_coords_2, y_coords_2),
            crs=source_raster.crs,
        )
        return gu.PointCloud(pc, data_column=data_column_name)
    else:
        # Merge the coordinates and pixel data an array of N x K
        # This has the downside of converting all the data to the same data type
        points_arr = np.vstack((x_coords_2.reshape(1, -1), y_coords_2.reshape(1, -1), pixel_data)).T
        return points_arr



####
# Part not shown in the doc, to get data files ready
import geoutils

landsat_b4_path = geoutils.examples.get_path("everest_landsat_b4")
everest_outlines_path = geoutils.examples.get_path("everest_rgi_outlines")
geoutils.Raster(landsat_b4_path).to_file("myraster.tif")
geoutils.Vector(everest_outlines_path).to_file("myvector.gpkg")
####

import geoutils as gu

# Opening a vector and a raster
vect = gu.Vector("myvector.gpkg")
rast = gu.Raster("myraster.tif")

# Metric buffering
vect_buff = vect.buffer_metric(buffer_size=100)

# Create a mask on the raster grid
# (raster not loaded, only metadata)
mask = vect_buff.create_mask(rast)

# Index raster values on mask
# (raster loads implicitly)
values = rast[mask]

####
# Part not shown in the doc, to get data files ready
import os

for file in ["myraster.tif", "myvector.gpkg"]:
    os.remove(file)
####



#!/usr/bin/env python
"""
Geoviewer provides a command line tool for plotting raster and vector data.

TO DO:
- include some options from imviewer: https://github.com/dshean/imview/blob/master/imview/imviewer.py
"""
from __future__ import annotations

import argparse
from typing import Sequence

import matplotlib.pyplot as plt
import numpy as np

from geoutils.raster import Raster


def getparser() -> argparse.ArgumentParser:
    # Set up description
    parser = argparse.ArgumentParser(
        description="Visualisation tool for any image supported by GDAL. For single band plots (Single band rasters or with option -band) the image will be rendered as a pseudocolor image using the set or default colormap. For 3 or 4 band data, the image will be plotted as an RGB(A) image. For other band counts, an error will be raised and the option -band must be used."
    )

    # Positional arguments
    parser.add_argument("filename", type=str, help="str, path to the image")

    # optional arguments
    parser.add_argument(
        "-cmap",
        dest="cmap",
        type=str,
        default="default",
        help="str, a matplotlib colormap string (default is from rcParams). This parameter is ignored for multi-band rasters.",
    )
    parser.add_argument(
        "-vmin",
        dest="vmin",
        type=str,
        default=None,
        help=(
            "float, the minimum value for colorscale, or can be expressed as a "
            "percentile e.g. 5%% (default is calculated min value). This parameter is ignored for multi-band rasters."
        ),
    )
    parser.add_argument(
        "-vmax",
        dest="vmax",
        type=str,
        default=None,
        help=(
            "float, the maximum value for colorscale, or can be expressed as a "
            "percentile e.g. 95%% (default is calculated max value). This parameter is ignored for multi-band rasters."
        ),
    )
    parser.add_argument(
        "-band",
        dest="band",
        type=int,
        default=None,
        help="int, for multiband images, which band to display. Starts at 1. (Default is to load all bands and display as rasterio, i.e. asuming RGB(A) and with clipping outside [0-255] for int, [0-1] for float).",
    )
    parser.add_argument(
        "-nocb",
        dest="nocb",
        help="If set, will not display a colorbar (Default is to display the colorbar for single-band raster). This parameter is ignored for multi-band rasters.",
        action="store_false",
    )
    parser.add_argument(
        "-clabel", dest="clabel", type=str, default="", help="str, the label for the colorscale (Default is empty)."
    )
    parser.add_argument("-title", dest="title", type=str, default="", help="str, figure title (Default is empty).")
    parser.add_argument(
        "-figsize",
        dest="figsize",
        type=str,
        default="default",
        help=(
            "str, figure size, must be a tuple of size 2, either written with quotes, "
            "or two numbers separated by comma, no space (Default is from rcParams)."
        ),
    )
    parser.add_argument(
        "-max_size",
        dest="max_size",
        type=int,
        default=2000,
        help="int, image size is limited to max_size**2 for faster reading/displaying (Default is 2000).",
    )
    parser.add_argument(
        "-save",
        dest="save",
        type=str,
        default="",
        help="str, filename to the output filename to save to disk (Default is displayed on screen).",
    )
    parser.add_argument(
        "-dpi",
        dest="dpi",
        type=str,
        default="default",
        help="int, dpi value to use when saving figure (Default is from rcParams).",
    )
    parser.add_argument(
        "-nodata",
        dest="nodata",
        type=str,
        default="default",
        help="float, no data value (Default is read from file metadata).",
    )
    parser.add_argument(
        "-noresampl",
        dest="noresampl",
        default=False,
        action="store_true",
        help="True or False, if False then allow dynamic image downscaling, if True, prevent it.",
    )

    return parser


def main(test_args: Sequence[str] = None) -> None:
    # Parse arguments
    parser = getparser()
    args = parser.parse_args(test_args)  # type: ignore

    # Load image metadata
    img = Raster(args.filename, load_data=False)

    # Resample if image is too large
    if ((img.width > args.max_size) or (img.height > args.max_size)) & (not args.noresampl):
        dfact = max(int(img.width / args.max_size), int(img.height / args.max_size))
        print(f"Image will be downsampled by a factor {dfact}.")
    else:
        dfact = 1

    # Read image
    img = Raster(args.filename, downsample=dfact, bands=args.band)

    # Set no data value
    if args.nodata == "default":
        pass
    else:
        try:
            nodata = float(args.nodata)
        except ValueError:
            raise ValueError("Nodata must be a float, currently set to %s" % args.nodata)

        img.set_nodata(nodata)

    # Set default parameters

    # vmin
    if args.vmin is not None:
        try:
            vmin: float | None = float(args.vmin)
        except ValueError:  # Case is not a number
            perc, _ = args.vmin.split("%")
            try:
                perc = float(perc)
                vmin = np.percentile(img.data.compressed(), perc)
            except ValueError:  # Case no % sign
                raise ValueError("vmin must be a float or percentage, currently set to %s" % args.vmin)

    else:
        vmin = None

    # vmax
    if args.vmax is not None:
        try:
            vmax: float | None = float(args.vmax)
        except ValueError:  # Case is not a number
            perc, _ = args.vmax.split("%")
            try:
                perc = float(perc)
                vmax = np.percentile(img.data.compressed(), perc)
            except ValueError:  # Case no % sign
                raise ValueError("vmax must be a float or percentage, currently set to %s" % args.vmax)

    else:
        vmax = None

    # color map
    # Get the list of existing color maps, including reversed maps
    mpl_cmap_list = list(plt.cm.datad.keys())
    mpl_cmap_list.extend([cmap + "_r" for cmap in mpl_cmap_list])

    if args.cmap == "default":
        cmap = plt.rcParams["image.cmap"]
    elif args.cmap in mpl_cmap_list:
        cmap = args.cmap
    else:
        raise ValueError("Wrong cmap, must be in: {}".format(",".join(str(elem) for elem in mpl_cmap_list)))

    # Figsize
    if args.figsize == "default":
        figsize = plt.rcParams["figure.figsize"]
    else:
        try:
            figsize = tuple(int(arg) for arg in args.figsize.split(","))
        except Exception:
            raise ValueError("Figsize must be a tuple of size 2, currently set to %s" % args.figsize)

    # dpi
    if args.dpi == "default":
        dpi = plt.rcParams["figure.dpi"]
    else:
        try:
            dpi = int(args.dpi)
        except ValueError:
            raise ValueError("dpi must be an integer, currently set to %s" % args.dpi)

    # Plot data

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    # plot
    img.plot(
        ax=ax,
        cmap=cmap,
        interpolation="nearest",
        vmin=vmin,
        vmax=vmax,
        add_cbar=args.nocb,
        cbar_title=args.clabel,
        title=args.title,
    )

    plt.tight_layout()

    # Save
    if args.save != "":
        plt.savefig(args.save, dpi=dpi)
        print("Figure saved to file %s." % args.save)
    else:
        plt.show()


if __name__ == "__main__":
    main()
